# Copyright (c) 2016-2020 The Regents of the University of Michigan
# This file is part of the General Simulation Data (GSD) project, released under
# the BSD 2-Clause License.

""" GSD reader written in pure python

:file:`pygsd.py` is a pure python implementation of a GSD reader. If your
analysis tool is written in python and you want to embed a GSD reader without
requiring C code compilation, then use the following python files from the
:file:`gsd/` directory to make a pure python reader. It is not as high
performance as the C reader, but is reasonable for files up to a few thousand
frames.

* :file:`gsd/`

    * :file:`__init__.py`
    * :file:`pygsd.py`
    * :file:`hoomd.py`


The reader reads from file-like python objects, which may be useful for reading
from in memory buffers, and in-database grid files, For regular files on the
filesystem, and for writing gsd files, use :py:mod:`gsd.fl`.

The :py:class:`GSDFile` in this module can be used with the
:py:class:`gsd.hoomd.HOOMDTrajectory` hoomd reader:

>>> with gsd.pygsd.GSDFile('test.gsd', 'rb') as f:
...     t = gsd.hoomd.HOOMDTrajectory(f)
...     pos = t[0].particles.position

"""

from __future__ import print_function
from __future__ import division
import logging
import numpy
import struct
from collections import namedtuple
import sys

__version__ = "2.1.0"

logger = logging.getLogger('gsd.pygsd')

gsd_header = namedtuple(
    'gsd_header',
    'magic index_location index_allocated_entries '
    'namelist_location namelist_allocated_entries '
    'schema_version gsd_version application '
    'schema reserved')
gsd_header_struct = struct.Struct('QQQQQII64s64s80s')

gsd_index_entry = namedtuple('gsd_index_entry',
                             'frame N location M id type flags')
gsd_index_entry_struct = struct.Struct('QQqIHBB')

gsd_type_mapping = {
    1: numpy.dtype('uint8'),
    2: numpy.dtype('uint16'),
    3: numpy.dtype('uint32'),
    4: numpy.dtype('uint64'),
    5: numpy.dtype('int8'),
    6: numpy.dtype('int16'),
    7: numpy.dtype('int32'),
    8: numpy.dtype('int64'),
    9: numpy.dtype('float32'),
    10: numpy.dtype('float64'), }


class GSDFile(object):
    """ GSDFile(file)

    GSD file access interface. Implemented in pure python and accepts any python
    file-like object.

    Args:

        file: File-like object to read.

    GSDFile implements an object oriented class interface to the GSD file
    layer. Use it to open an existing file in a **read-only** mode. For
    read-write access to files, use the full featured C implementation in
    :py:mod:`gsd.fl`. Otherwise, this implementation has all the same methods
    and the two classes can be used interchangeably.

    Examples:

        Open a file in **read-only** mode::

            f = GSDFile(open('file.gsd', mode='rb'))
            if f.chunk_exists(frame=0, name='chunk'):
                data = f.read_chunk(frame=0, name='chunk')

        Access file **metadata**::

            f = GSDFile(open('file.gsd', mode='rb'))
            print(f.name, f.mode, f.gsd_version)
            print(f.application, f.schema, f.schema_version)
            print(f.nframes)

        Use as a **context manager**::

            with GSDFile(open('file.gsd', mode='rb')) as f:
                data = f.read_chunk(frame=0, name='chunk')

    Attributes:

        file: File-like object opened **(read only)**.
        name (str): file.name **(read only)**.
        mode (str): Mode of the open file **(read only)**.
        gsd_version (tuple[int]): GSD file layer version number [major, minor]
          **(read only)**.
        application (str): Name of the generating application **(read only)**.
        schema (str): Name of the data schema **(read only)**.
        schema_version (tuple[int]): Schema version number [major, minor]
          **(read only)**.
        nframes (int): Number of frames **(read only)**.
    """
    def __init__(self, file):
        self.__file = file

        logger.info('opening file: ' + str(file))

        # read the header
        self.__file.seek(0)
        try:
            header_raw = self.__file.read(gsd_header_struct.size)
        except UnicodeDecodeError:
            print("\nDid you open the file in binary mode (rb)?\n",
                  file=sys.stderr)
            raise

        if len(header_raw) != gsd_header_struct.size:
            raise IOError

        self.__header = gsd_header._make(gsd_header_struct.unpack(header_raw))

        # validate the header
        if self.__header.magic != 0x65DF65DF65DF65DF:
            raise RuntimeError("Not a GSD file: " + str(self.__file))
        if (self.__header.gsd_version < (1 << 16)
                and self.__header.gsd_version != (0 << 16 | 3)):
            raise RuntimeError("Unsupported GSD file version: "
                               + str(self.__file))
        if self.__header.gsd_version >= (3 << 16):
            raise RuntimeError("Unsupported GSD file version: "
                               + str(self.__file))

        # determine the file size (only works in python 3)
        self.__file.seek(0, 2)

        # read the namelist block into a dict for easy lookup
        self.__namelist = {}
        c = 0
        self.__file.seek(self.__header.namelist_location, 0)
        namelist_raw = self.__file.read(self.__header.namelist_allocated_entries
                                        * 64)

        names = namelist_raw.split(b'\x00')

        for name in names:
            sname = name.decode('utf-8')
            if len(sname) != 0:
                self.__namelist[sname] = c
                c = c + 1

        # read the index block. Since this is a read-only implementation, only
        # read in the used entries
        self.__index = []
        self.__file.seek(self.__header.index_location, 0)
        for i in range(self.__header.index_allocated_entries):
            index_entry_raw = self.__file.read(gsd_index_entry_struct.size)
            if len(index_entry_raw) != gsd_index_entry_struct.size:
                raise IOError

            idx = gsd_index_entry._make(
                gsd_index_entry_struct.unpack(index_entry_raw))

            # 0 location signifies end of index
            if idx.location == 0:
                break

            if not self.__is_entry_valid(idx):
                raise RuntimeError("Corrupt GSD file: " + str(self.__file))

            if i > 0 and idx.frame < self.__index[i - 1].frame:
                raise RuntimeError("Corrupt GSD file: " + str(self.__file))

            self.__index.append(idx)

        self.__is_open = True

    def __is_entry_valid(self, entry):
        """ Return True if an entry is valid
        """
        if entry.type not in gsd_type_mapping:
            return False

        if entry.M == 0:
            return False

        if entry.frame >= self.__header.index_allocated_entries:
            return False

        if entry.id >= len(self.__namelist):
            return False

        if entry.flags != 0:
            return False

        return True

    def close(self):
        """ close()

        Close the file.

        Once closed, any other operation on the file object will result in a
        `ValueError`. :py:meth:`close()` may be called more than once.
        The file is automatically closed when garbage collected or when
        the context manager exits.
        """
        if self.__is_open:
            logger.info('closing file: ' + str(self.__file))
            self.__handle = None
            self.__index = None
            self.__namelist = None
            self.__is_open = False
            self.__file.close()

    def truncate(self):
        raise NotImplementedError

    def end_frame(self):
        raise NotImplementedError

    def write_chunk(self, name, data):
        raise NotImplementedError

    def _find_chunk(self, frame, name):
        # find the id for the given name
        if name in self.__namelist:
            match_id = self.__namelist[name]
        else:
            return None

        # TODO: optimize for v2.0 files
        # binary search for the first index entry at the requested frame
        L = 0
        R = len(self.__index)

        # progressively narrow the search window by halves
        while (R - L > 1):
            m = (L + R) // 2

            if frame < self.__index[m].frame:
                R = m
            else:
                L = m

        # this finds L = the rightmost index with the desired frame
        # search all index entries with the matching frame
        cur_index = L
        while cur_index >= 0 and self.__index[cur_index].frame == frame:
            if match_id == self.__index[cur_index].id:
                return self.__index[cur_index]
            cur_index = cur_index - 1

        # if we got here, we didn't find the specified chunk
        return None

    def chunk_exists(self, frame, name):
        """ chunk_exists(frame, name)

        Test if a chunk exists.

        Args:
            frame (int): Index of the frame to check
            name (str): Name of the chunk

        Returns:
            bool: True if the chunk exists in the file. False if it does not.

        Example:

            Handle non-existent chunks::

                with GSDFile(open('file.gsd', mode='rb')) as f:
                    if f.chunk_exists(frame=0, name='chunk'):
                        return f.read_chunk(frame=0, name='chunk')
                    else:
                        return None
        """

        if not self.__is_open:
            raise ValueError("File is not open")

        chunk = self._find_chunk(frame, name)
        return chunk is not None

    def read_chunk(self, frame, name):
        """ read_chunk(frame, name)

        Read a data chunk from the file and return it as a numpy array.

        Args:
            frame (int): Index of the frame to read
            name (str): Name of the chunk

        Returns:
            ``numpy.ndarray[type, ndim=?, mode='c']``: Data read from file.
              ``type`` is determined by the chunk metadata. If the data is
              NxM in the file and M > 1, return a 2D array. If the data is
              Nx1, return a 1D array.

        Examples:

            Read a 1D array::

                with GSDFile(name=filename, mode='rb') as f:
                    data = f.read_chunk(frame=0, name='chunk1d')
                    # data.shape == [N]

            Read a 2D array::

                with GSDFile(name=filename, mode='rb') as f:
                    data = f.read_chunk(frame=0, name='chunk2d')
                    # data.shape == [N,M]

            Read multiple frames::

                with GSDFile(name=filename, mode='rb') as f:
                    data0 = f.read_chunk(frame=0, name='chunk')
                    data1 = f.read_chunk(frame=1, name='chunk')
                    data2 = f.read_chunk(frame=2, name='chunk')
                    data3 = f.read_chunk(frame=3, name='chunk')

        .. tip::
            Each call invokes a disk read and allocation of a
            new numpy array for storage. To avoid overhead, don't call
            :py:meth:`read_chunk()` on the same chunk repeatedly. Cache the
            arrays instead.
        """

        if not self.__is_open:
            raise ValueError("File is not open")

        chunk = self._find_chunk(frame, name)

        if chunk is None:
            raise KeyError("frame " + str(frame) + " / chunk " + name
                           + " not found in: " + str(self.__file))

        logger.debug('read chunk: ' + str(self.__file) + ' - ' + str(frame)
                     + ' - ' + name)

        size = chunk.N * chunk.M * gsd_type_mapping[chunk.type].itemsize
        if chunk.location == 0:
            raise RuntimeError("Corrupt chunk: " + str(frame) + " / " + name
                               + " in file" + str(self.__file))

        if (size == 0):
            return numpy.array([], dtype=gsd_type_mapping[chunk.type])

        self.__file.seek(chunk.location, 0)
        data_raw = self.__file.read(size)

        if len(data_raw) != size:
            raise IOError

        data_npy = numpy.frombuffer(data_raw,
                                    dtype=gsd_type_mapping[chunk.type])

        if chunk.M == 1:
            return data_npy
        else:
            return data_npy.reshape([chunk.N, chunk.M])

    def find_matching_chunk_names(self, match):
        """ find_matching_chunk_names(match)

        Find all the chunk names in the file that start with the string *match*.

        Args:
            match (str): Start of the chunk name to match

        Returns:
            list[str]: Matching chunk names
        """
        result = []
        for key in self.__namelist.keys():
            if key.startswith(match):
                result.append(key)

        return result

    def __getstate__(self):
        return dict(name=self.name)

    def __setstate__(self, state):
        self.__init__(open(state['name'], 'rb'))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def name(self):
        return self.__file.name

    @property
    def file(self):
        return self.__file

    @property
    def mode(self):
        return 'rb'

    @property
    def gsd_version(self):
        v = self.__header.gsd_version
        return (v >> 16, v & 0xffff)

    @property
    def schema_version(self):
        v = self.__header.schema_version
        return (v >> 16, v & 0xffff)

    @property
    def schema(self):
        return self.__header.schema.rstrip(b'\x00').decode('utf-8')

    @property
    def application(self):
        return self.__header.application.rstrip(b'\x00').decode('utf-8')

    @property
    def nframes(self):
        if not self.__is_open:
            raise ValueError("File is not open")

        if len(self.__index) == 0:
            return 0
        else:
            return self.__index[-1].frame + 1



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.


__version__ = "0.7.1"


__all__ = [
    '__version__'
]



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""POS-file writer for the Glotzer Group, University of Michigan.

Author: Carl Simon Adorf

.. code::

    writer = PosFileWriter()
    with open('a_posfile.pos', 'w', encoding='utf-8') as posfile:
        writer.write(trajectory, posfile)
"""

import io
import sys
import logging
import warnings
import math
from itertools import chain

import numpy as np

from .posfilereader import POSFILE_FLOAT_DIGITS
from .shapes import SphereShape, ArrowShape
import rowan


logger = logging.getLogger(__name__)

DEFAULT_SHAPE_DEFINITION = SphereShape(1.0, color='005984FF')


def _num(x):
    "Round x if x is a floating point number."
    return int(x) if int(x) == x else round(float(x), POSFILE_FLOAT_DIGITS)


class PosFileWriter(object):
    """POS-file writer for the Glotzer Group, University of Michigan.

    Author: Carl Simon Adorf

    .. code::

        writer = PosFileWriter()
        with open('a_posfile.pos', 'w', encoding='utf-8') as posfile:
            writer.write(trajectory, posfile)

    :param rotate: Rotate the system into the view rotation instead of adding
        it to the metadata with the 'rotation' keyword.
    :type rotate: bool
    """
    def __init__(self, rotate=False):
        self._rotate = rotate
        if self._rotate:
            warnings.warn(
                "Rotating the system with a view rotation leads to significant "
                "numerical precision loss!")

    def write(self, trajectory, file=sys.stdout):
        """Serialize a trajectory into pos-format and write it to file.

        :param trajectory: The trajectory to serialize
        :type trajectory: :class:`~garnett.trajectory.Trajectory`
        :param file: A file-like object."""
        def _write(msg, end='\n'):
            file.write(msg + end)
        for i, frame in enumerate(trajectory):
            # data section
            if frame.data is not None:
                header_keys = frame.data_keys
                _write('#[data] ', end='')
                _write(' '.join(header_keys))
                columns = list()
                for key in header_keys:
                    columns.append(frame.data[key])
                rows = np.array(columns).transpose()
                for row in rows:
                    _write(' '.join(row))
                _write('#[done]')

            # boxMatrix and rotation
            box_matrix = np.array(frame.box.get_box_matrix())
            if self._rotate and frame.view_rotation is not None:
                for i in range(3):
                    box_matrix[:, i] = rowan.rotate(frame.view_rotation, box_matrix[:, i])

            if frame.view_rotation is not None and not self._rotate:
                angles = rowan.to_euler(frame.view_rotation, axis_type='extrinsic', convention='xyz') * 180 / math.pi
                _write('rotation ' + ' '.join((str(_num(_)) for _ in angles)))

            _write('boxMatrix ', end='')
            _write(' '.join((str(_num(v)) for v in box_matrix.flatten())))

            # shape defs
            try:
                if len(frame.types) != len(frame.type_shapes):
                    raise ValueError("Unequal number of types and type_shapes.")
                for name, type_shape in zip(frame.types, frame.type_shapes):
                    _write('def {} "{}"'.format(name, type_shape.pos_string))
            except AttributeError:
                # If AttributeError is raised because the frame does not contain
                # shape information, fill them all with the default shape
                for name in frame.types:
                    logger.info(
                        "No shape defined for '{}'. "
                        "Using fallback definition.".format(name))
                    _write('def {} "{}"'.format(name, DEFAULT_SHAPE_DEFINITION.pos_string))

            # Orientations must be provided for all particles
            # If the frame does not have orientations, identity quaternions are used
            orientation = getattr(frame, 'orientation', np.array([[1, 0, 0, 0]] * frame.N))

            for typeid, pos, rot in zip(frame.typeid, frame.position, orientation):
                name = frame.types[typeid]
                _write(name, end=' ')
                try:
                    shapedef = frame.shapedef.get(name)
                except AttributeError:
                    shapedef = DEFAULT_SHAPE_DEFINITION

                if self._rotate and frame.view_rotation is not None:
                    pos = rowan.rotate(frame.view_rotation, pos)
                    rot = rowan.multiply(frame.view_rotation, rot)

                if isinstance(shapedef, SphereShape):
                    _write(' '.join((str(_num(v)) for v in pos)))
                elif isinstance(shapedef, ArrowShape):
                    # The arrow shape actually has two position vectors of
                    # three elements since it has start.{x,y,z} and end.{x,y,z}.
                    # That is, "rot" is not an accurate variable name, since it
                    # does not represent a quaternion.
                    _write(' '.join((str(_num(v)) for v in chain(pos, rot[:3]))))
                else:
                    _write(' '.join((str(_num(v)) for v in chain(pos, rot))))
            _write('eof')
            logger.debug("Wrote frame {}.".format(i + 1))
        logger.info("Wrote {} frames.".format(i + 1))

    def dump(self, trajectory):
        """Serialize trajectory into pos-format.

        :param trajectory: The trajectory to serialize.
        :type trajectory: :class:`~garnett.trajectory.Trajectory`
        :rtype: str"""
        f = io.StringIO()
        self.write(trajectory, f)
        return f.getvalue()



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""CIF-file reader for the Glotzer Group, University of Michigan.

Authors: Matthew Spellings

.. code::

    reader = CifFileReader()
    with open('a_ciffile.cif', 'r', encoding='utf-8') as ciffile:
        traj = reader.read(ciffile)
"""

import logging
import re
import warnings
from collections import OrderedDict

import numpy as np

from .trajectory import _RawFrameData, Frame, Trajectory

# CifFile is from the pycifrw package
from CifFile import CifFile

from .errors import ParserError, ParserWarning

logger = logging.getLogger(__name__)

CIFFILE_FLOAT_DIGITS = 11

# invalid characters for symmetry expressions
REMOVE_NONNUM_REGEXP = re.compile(r'[^+\-*/0-9\.,\(\)xyz]+')

# (an integer or floating point number)/(an integer or floating point number)
PARSE_DIVISION_REGEXP = re.compile(r'(?P<num>\d+(\.(\d+)?)?)\s*/\s*(?P<denom>\d+(\.(\d+)?)?)')

# This regex matches _atom_site_label component 0 (but not component 1) of the
# definition here:
# https://www.iucr.org/__data/iucr/cif/standard/cifstd15.html
_ATOM_SITE_LABEL_COMPONENT_0 = re.compile(r'^(([^\s_0-9])|([0-9]*[\+\-]))*')


def _parse_division(match):
    """Helper function to substitute ratios in cif files, like '1/4', into
    fractions, like '0.25'."""
    return str(float(match.group('num'))/float(match.group('denom')))


def _parse_atom_site_label_to_type_name(label):
    """Matches _atom_site_label component 0.

    This regex matches component 0 (but not component 1) of the definition here:
    https://www.iucr.org/__data/iucr/cif/standard/cifstd15.html
    """
    m = re.search(_ATOM_SITE_LABEL_COMPONENT_0, label)
    return m.group()


class _RawCifFrameData(_RawFrameData):
    """Extend base class to support raw CIF coordinates"""

    def __init__(self):
        # Append the cif_coordinates to the dataset
        super(_RawCifFrameData, self).__init__()
        self.cif_coordinates = list()               # Nx3


class CifFileFrame(Frame):

    def __init__(self, parsed, precision, default_type, tolerance=1e-5):
        self.parsed = parsed
        self.precision = precision
        self.default_type = default_type
        self.tolerance = tolerance
        super(CifFileFrame, self).__init__()

    def _num(self, x):
        if isinstance(x, int):
            return x
        else:
            return round(float(x), self.precision)

    def _safer_float(self, s):
        """Only takes the first component of a string as a number, in case
        some notion of imprecision is specified e.g. 1.3141(2)"""
        idx = s.rfind('(')
        return self._num(s[:idx]) if idx >= 0 else self._num(s)

    def _parse_box(self, parsed):
        (a, b, c) = [self._safer_float(parsed['_cell_length_{}'.format(name)])
                     for name in ['a', 'b', 'c']]
        (alpha, beta, gamma) = [self._safer_float(parsed['_cell_angle_{}'.format(name)])
                                for name in ['alpha', 'beta', 'gamma']]
        alpha *= np.pi/180
        beta *= np.pi/180
        gamma *= np.pi/180

        # http://lammps.sandia.gov/doc/Section_howto.html#howto-12
        # in this block, use the definitions lammps uses
        # (lammps xy == hoomd xy*ly, etc)
        lx = a
        xy = b*np.cos(gamma)
        xz = c*np.cos(beta)
        ly = np.sqrt(b**2 - xy**2)
        yz = (b*c*np.cos(alpha) - xy*xz)/ly
        lz = np.sqrt(c**2 - xz**2 - yz**2)

        return np.array([[lx, xy, xz],
                         [0, ly, yz],
                         [0, 0, lz]])

    @property
    def cif_coordinates(self):
        "Nx3 matrix of exact coordinates provided in the CIF file."
        self.load()
        return self._frame_data.cif_coordinates

    @cif_coordinates.setter
    def cif_coordinates(self, value):
        try:
            value = np.asarray(value, dtype=self._dtype)
        except ValueError:
            raise ValueError("CIF coordinates can only be set to numeric arrays.")
        if not np.all(np.isfinite(value)):
            raise ValueError("CIF coordinates being set must all be finite numbers.")
        elif not len(value.shape) == 2 or value.shape[1] != self.box.dimensions:
            raise ValueError("Input array must be of shape (N,{}) where N is the "
                             "number of particles.".format(self.box.dimensions))

        self.load()
        self._frame_data.cif_coordinates = value

    @staticmethod
    def _raw_frame_data_to_frame_data(raw_frame, dtype=None):
        """Extend parent function to also incorporate cif_coordinates"""
        ret = super(CifFileFrame, CifFileFrame)._raw_frame_data_to_frame_data(raw_frame, dtype)
        ret.cif_coordinates = np.asarray(raw_frame.cif_coordinates, dtype=dtype)
        assert len(ret.position) == len(ret.cif_coordinates)
        return ret

    def read(self):
        "Read the frame data from the stream."
        box_matrix = self._parse_box(self.parsed)

        fractions = np.array([(self._safer_float(x), self._safer_float(y), self._safer_float(z))
                              for (x, y, z) in zip(
            self.parsed['_atom_site_fract_x'], self.parsed['_atom_site_fract_y'], self.parsed['_atom_site_fract_z'])])

        if '_atom_site_type_symbol' in self.parsed:
            site_types = list(self.parsed['_atom_site_type_symbol'])
        elif '_atom_site_label' in self.parsed:
            site_types = [_parse_atom_site_label_to_type_name(label)
                          for label in self.parsed['_atom_site_label']]
        else:
            site_types = len(fractions)*[self.default_type]

        space_group_keys = ['_symmetry_equiv_pos_as_xyz', '_space_group_symop_operation_xyz']
        found_keys = [key for key in space_group_keys if key in self.parsed]

        unique_points = []
        type_strings = []

        if found_keys:
            key_to_use = found_keys[0]
            symmetry_ops = [PARSE_DIVISION_REGEXP.sub(
                            _parse_division, REMOVE_NONNUM_REGEXP.sub('', sym))
                            for sym in self.parsed[key_to_use]]

            replicated_fractions = []
            replicated_types = []
            for (type_name, (fx, fy, fz)) in zip(site_types, fractions):
                extra_fractions = [eval(sym, dict(x=fx, y=fy, z=fz)) for sym in symmetry_ops]
                replicated_fractions.extend(extra_fractions)
                replicated_types.extend(len(extra_fractions)*[type_name])
            # wrap back into the box
            replicated_fractions -= np.floor(replicated_fractions)
            replicated_fractions = dict(enumerate(np.array(replicated_fractions, dtype=np.float32)))

            bad_types = False
            # short of using scipy or freud, we just exhaustively search
            # for points near each point to find symmetry-induced duplicates
            while len(replicated_fractions):
                (ref_index, ref_point) = replicated_fractions.popitem()
                type_name = replicated_types[ref_index]

                # set of points that are ~equivalent to ref_point given a tolerance
                current_points = [ref_point]

                # find similar points and add them to the list
                for index in list(replicated_fractions):
                    # find nearest periodic image
                    delta = replicated_fractions[index] - ref_point
                    delta = ((delta + 0.5) % 1.0) - 0.5

                    if np.allclose(delta, 0, atol=self.tolerance):
                        del replicated_fractions[index]
                        current_points.append(ref_point + delta)

                        if replicated_types[ref_index] != replicated_types[index]:
                            bad_types = True
                            msg = ('Some distinct sites were merged into the same '
                                   'position, the types for this file are invalid.')
                            warnings.warn(msg, ParserWarning)

                type_strings.append(type_name)
                unique_points.append(np.mean(current_points, axis=0))
            unique_points = np.array(unique_points, dtype=np.float32)

            if bad_types:
                site_types = [self.default_type]
                type_strings = [self.default_type] * len(unique_points)
        else:
            unique_points = fractions
            type_strings = site_types

        # Remove duplicate site_types while preserving their order
        site_types = list(OrderedDict.fromkeys(site_types))

        # Convert type strings to typeid
        site_types_lookup = {key: index for index, key in enumerate(site_types)}
        typeid = [site_types_lookup[t] for t in type_strings]

        # Save the exact points
        cif_coordinates = unique_points.copy()

        # shift so that (0, 0, 0) in fractional coordinates goes to a
        # corner of the box, not the center of the box
        unique_points -= 0.5

        coordinates = np.sum(
            unique_points[:, np.newaxis, :]*box_matrix[np.newaxis, :, :], axis=2)

        raw_frame = _RawFrameData()
        raw_frame.box = box_matrix
        raw_frame.types = site_types
        raw_frame.typeid = typeid
        raw_frame.position = coordinates
        raw_frame.cif_coordinates = cif_coordinates
        return raw_frame

    def __str__(self):
        return "CifFileFrame(parsed={}, precision={})".format(
            self.parsed, self.precision)


class CifFileReader(object):
    """CIF-file reader for the Glotzer Group, University of Michigan.

        Requires the PyCifRW package to parse CIF files.

        Authors: Matthew Spellings

        .. code::

            reader = CifFileReader()
            with open('a_ciffile.cif', 'r') as ciffile:
                traj = reader.read(ciffile)

        :param precision: The number of digits to
                          round floating-point values to.
        :type precision: int
        :param tolerance: Floating-point tolerance (in fractional
                          coordinates) of particle identity as symmetry
                          operations are applied.
        :type tolerance: float
    """

    def __init__(self, precision=None, tolerance=1e-3):
        """Initialize a cif-file reader."""
        self._precision = precision or CIFFILE_FLOAT_DIGITS
        self._tolerance = tolerance

    def _scan(self, parsed_file, keys, default_type):
        return (CifFileFrame(parsed_file[key], self._precision, default_type,
                             self._tolerance) for key in keys)

    def read(self, stream, default_type='A'):
        """Read text stream and return a trajectory instance.

        :param stream: The stream, which contains the ciffile.
        :type stream: A file-like textstream.
        :param default_type: The default particle type for
                             ciffile dialects without type definition.
        :type default_type: str
        """
        parsed_file = CifFile(stream)
        keys = list(sorted(parsed_file.keys()))

        # Index the stream
        frames = list(self._scan(parsed_file, keys, default_type))
        if len(frames) == 0:
            raise ParserError("Did not read a single complete frame.")
        logger.info("Read {} frames.".format(len(frames)))
        return Trajectory(frames)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""HOOMD-blue GSD file reader for the Glotzer Group, University of Michigan.

Author: Carl Simon Adorf

This module provides a wrapper for the trajectory reader
implementation as part of the gsd package.

A gsd file may not contain all shape information.
To provide additional information it is possible
to pass a frame object, whose properties
are copied into each frame of the gsd trajectory.

This example reads shape data from a HOOMD-blue POS frame:

.. code::

    pos_reader = PosFileReader()
    gsd_reader = GSDHOOMDFileReader()

    with open('init.pos') as posfile:
        with open('dump.gsd') as gsdfile:
            pos_frame = pos_reader.read(posfile)[0]
            traj = gsd_reader.read(gsdfile, pos_frame)
"""

import logging
import warnings
import copy

import numpy as np

from .trajectory import _RawFrameData, Frame, Trajectory
from .shapes import SphereShape, ConvexPolyhedronShape, ConvexSpheropolyhedronShape, \
    PolygonShape, SpheropolygonShape, EllipsoidShape, _parse_type_shape

try:
    import gsd
    NATIVE = True
except ImportError:
    NATIVE = False
from .pygsd import GSDFile as PyGSDFile
from . import gsdhoomd


logger = logging.getLogger(__name__)


def _box_matrix(box):
    lx, ly, lz, xy, xz, yz = box
    return np.array([
        [lx, 0.0, 0.0],
        [xy * ly, ly, 0.0],
        [xz * lz, yz * lz, lz]]).T


def _parse_shape_definitions(frame, gsdfile, frame_index):
    """Parses shape information for a frame in a GSD file.

    This method uses the ``type_shapes`` property if it is set. Otherwise it
    falls back to reading HPMC state data and inferring ``type_shapes``.
    """

    def get_chunk(i, chunk, default=None):
        if gsdfile.chunk_exists(i, chunk):
            return gsdfile.read_chunk(i, chunk)
        elif gsdfile.chunk_exists(0, chunk):
            return gsdfile.read_chunk(0, chunk)
        else:
            return default

    if get_chunk(frame_index, 'particles/type_shapes') is not None:
        # --------------------------------------------------
        # Parsing from GSD Shape Visualization Specification
        # --------------------------------------------------
        return [_parse_type_shape(t) for t in frame.particles.type_shapes]
    else:
        # -----------------------
        # Parsing from HPMC State
        # -----------------------

        type_shapes = []
        # Spheres
        if get_chunk(frame_index, 'state/hpmc/sphere/radius') is not None:
            radii = get_chunk(frame_index, 'state/hpmc/sphere/radius')
            orientables = get_chunk(frame_index, 'state/hpmc/sphere/orientable')
            # Since the orientable chunk was only added in HOOMD Schema version 1.3,
            # not all GSD files may have it. Thus, it is set to False in such cases.
            if orientables is None:
                orientables = [False]*len(radii)
            for radius, orientable in zip(radii, orientables):
                shape = SphereShape(
                    diameter=radius*2, orientable=orientable, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Convex Polyhedra
        elif get_chunk(frame_index, 'state/hpmc/convex_polyhedron/N') is not None:
            N = get_chunk(frame_index, 'state/hpmc/convex_polyhedron/N')
            N_start = [sum(N[:i]) for i in range(len(N))]
            N_end = [sum(N[:i+1]) for i in range(len(N))]
            verts = get_chunk(frame_index, 'state/hpmc/convex_polyhedron/vertices')
            verts_split = [verts[start:end] for start, end in zip(N_start, N_end)]
            for typeverts in verts_split:
                shape = ConvexPolyhedronShape(
                    vertices=typeverts, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Convex Spheropolyhedra
        elif get_chunk(frame_index, 'state/hpmc/convex_spheropolyhedron/N') is not None:
            N = get_chunk(frame_index, 'state/hpmc/convex_spheropolyhedron/N')
            N_start = [sum(N[:i]) for i in range(len(N))]
            N_end = [sum(N[:i+1]) for i in range(len(N))]
            verts = get_chunk(frame_index,
                              'state/hpmc/convex_spheropolyhedron/vertices')
            verts_split = [verts[start:end] for start, end in zip(N_start, N_end)]
            sweep_radii = get_chunk(frame_index,
                                    'state/hpmc/convex_spheropolyhedron/sweep_radius')
            for typeverts, radius in zip(verts_split, sweep_radii):
                shape = ConvexSpheropolyhedronShape(
                    vertices=typeverts, rounding_radius=radius, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Ellipsoid
        elif get_chunk(frame_index, 'state/hpmc/ellipsoid/a') is not None:
            a_all = get_chunk(frame_index, 'state/hpmc/ellipsoid/a')
            b_all = get_chunk(frame_index, 'state/hpmc/ellipsoid/b')
            c_all = get_chunk(frame_index, 'state/hpmc/ellipsoid/c')
            for a, b, c in zip(a_all, b_all, c_all):
                shape = EllipsoidShape(
                    a=a, b=b, c=c, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Convex Polygons
        elif get_chunk(frame_index, 'state/hpmc/convex_polygon/N') is not None:
            N = get_chunk(frame_index, 'state/hpmc/convex_polygon/N')
            N_start = [sum(N[:i]) for i in range(len(N))]
            N_end = [sum(N[:i+1]) for i in range(len(N))]
            verts = get_chunk(frame_index, 'state/hpmc/convex_polygon/vertices')
            verts_split = [verts[start:end] for start, end in zip(N_start, N_end)]
            for typeverts in verts_split:
                shape = PolygonShape(
                    vertices=typeverts, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Convex Spheropolygons
        elif get_chunk(frame_index, 'state/hpmc/convex_spheropolygon/N') is not None:
            N = get_chunk(frame_index, 'state/hpmc/convex_spheropolygon/N')
            N_start = [sum(N[:i]) for i in range(len(N))]
            N_end = [sum(N[:i+1]) for i in range(len(N))]
            verts = get_chunk(frame_index, 'state/hpmc/convex_spheropolygon/vertices')
            verts_split = [verts[start:end] for start, end in zip(N_start, N_end)]
            sweep_radii = get_chunk(frame_index,
                                    'state/hpmc/convex_spheropolygon/sweep_radius')
            for typeverts, radius in zip(verts_split, sweep_radii):
                shape = SpheropolygonShape(
                    vertices=typeverts, rounding_radius=radius, color=None)
                type_shapes.append(shape)
            return type_shapes

        # Simple Polygons
        elif get_chunk(frame_index, 'state/hpmc/simple_polygon/N') is not None:
            N = get_chunk(frame_index, 'state/hpmc/simple_polygon/N')
            N_start = [sum(N[:i]) for i in range(len(N))]
            N_end = [sum(N[:i+1]) for i in range(len(N))]
            verts = get_chunk(frame_index, 'state/hpmc/simple_polygon/vertices')
            verts_split = [verts[start:end] for start, end in zip(N_start, N_end)]
            for typeverts in verts_split:
                shape = PolygonShape(
                    vertices=typeverts, color=None)
                type_shapes.append(shape)
            return type_shapes

        # If no shapes were detected, return None
        else:
            return None


class GSDHoomdFrame(Frame):
    """Extends the Frame object for GSD files.

    :param traj:
        Trajectory containing the frame to cast
    :type traj:
        :class:`trajectory.Trajectory`
    :param frame_index:
        The index of the frame to cast
    :type frame_index:
        int
    :param t_frame:
        A frame containing shape information that is not encoded in
        the GSD-format. By default, shape information is read from the
        passed frame object, if one provided. Otherwise, shape information
        is read from the gsd file.
    :type :
        :class:`trajectory.Frame`
    :param gsdfile:
        A gsd file object.
    :type gsdfile:
        :class:`gsd.fl.GSDFile`
    """

    def __init__(self, traj, frame_index, t_frame, gsdfile):
        self.traj = traj
        self.frame_index = frame_index
        self.t_frame = t_frame
        self.gsdfile = gsdfile
        super(GSDHoomdFrame, self).__init__()

    def read(self):
        raw_frame = _RawFrameData()
        frame = self.traj.read_frame(self.frame_index)
        # If frame is provided, read shape data from it
        if self.t_frame is not None:
            raw_frame.data = copy.deepcopy(self.t_frame.data)
            raw_frame.data_keys = copy.deepcopy(self.t_frame.data_keys)
            raw_frame.box_dimensions = self.t_frame.box.dimensions
            try:
                raw_frame.type_shapes = copy.deepcopy(self.t_frame.type_shapes)
            except AttributeError:
                pass
        else:
            # Fallback to gsd shape data if no frame is provided
            raw_frame.type_shapes = _parse_shape_definitions(frame, self.gsdfile, self.frame_index)
        raw_frame.box = _box_matrix(frame.configuration.box)
        raw_frame.box_dimensions = int(frame.configuration.dimensions)
        raw_frame.types = frame.particles.types
        raw_frame.typeid = frame.particles.typeid
        raw_frame.position = frame.particles.position
        raw_frame.orientation = frame.particles.orientation
        raw_frame.velocity = frame.particles.velocity
        raw_frame.mass = frame.particles.mass
        raw_frame.charge = frame.particles.charge
        raw_frame.diameter = frame.particles.diameter
        raw_frame.moment_inertia = frame.particles.moment_inertia
        raw_frame.angmom = frame.particles.angmom
        raw_frame.image = frame.particles.image
        return raw_frame

    def __str__(self):
        return "GSDHoomdFrame(# frames={})".format(len(self.traj))


class GSDHOOMDFileReader(object):
    """HOOMD-blue GSD file reader for the Glotzer Group, University of Michigan.

    Author: Carl Simon Adorf

    This class provides a wrapper for the trajectory reader
    implementation as part of the gsd package.

    A gsd file may not contain all shape information.
    To provide additional information it is possible
    to pass a frame object, whose properties
    are copied into each frame of the gsd trajectory.

    The example is given for a HOOMD-blue XML frame:

    .. code::

        xml_reader = HOOMDXMLFileReader()
        gsd_reader = GSDHOOMDFileReader()

        with open('init.xml') as xmlfile:
            with open('dump.gsd', 'rb') as gsdfile:
                xml_frame = xml_reader.read(xmlfile)[0]
                traj = gsd_reader.read(gsdfile, xml_frame)

    About the read_gsd_shape_data parameter: This parameter was removed. By default,
                shape information is read from a passed frame object, if one
                provided. Otherwise, shape information is read from the gsd file.

    """

    def __init__(self):
        pass

    def read(self, stream, frame=None):
        """Read binary stream and return a trajectory instance.

        :param stream: The stream, which contains the gsd-file.
        :type stream: A file-like binary stream
        :param frame: A frame containing shape information
            that is not encoded in the GSD-format. By default,
            shape information is read from the passed frame object,
            if one provided. Otherwise, shape information
            is read from the gsd file.
        :type frame: :class:`trajectory.Frame`"""
        if NATIVE:
            try:
                traj = gsd.hoomd.open(name=stream.name, mode="rb")
                gsdfile = traj.file
            except AttributeError:
                logger.info(
                    "Unable to open file stream natively, falling back "
                    "to pure python GSD reader.")
                gsdfile = PyGSDFile(stream)
                traj = gsdhoomd.HOOMDTrajectory(gsdfile)
        else:
            warnings.warn("Native GSD library not available. "
                          "Falling back to pure python reader.")
            gsdfile = PyGSDFile(stream)
            traj = gsdhoomd.HOOMDTrajectory(gsdfile)
        frames = [GSDHoomdFrame(traj, i, t_frame=frame, gsdfile=gsdfile)
                  for i in range(len(traj))]
        logger.info("Read {} frames.".format(len(frames)))
        return Trajectory(frames)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.


class ParserError(RuntimeError):
    pass


class ParserWarning(RuntimeWarning):
    pass



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

"""Utility functions for format detection and simple reading/writing."""
import os
import logging
from contextlib import contextmanager
from . import reader, writer


logger = logging.getLogger(__name__)


# Mapping of file extension to format
FORMATS = {
    '.pos': 'pos',
    '.gsd': 'gsd',
    '.zip': 'gtar',
    '.tar': 'gtar',
    '.sqlite': 'gtar',
    '.dcd': 'dcd',
    '.xml': 'xml',
    '.cif': 'cif'}

# Mapping of format to reader classes and input modes
READ_CLASS_MODES = {
    'pos': {
        'reader': reader.PosFileReader,
        'mode': 'r'},
    'gsd': {
        'reader': reader.GSDHOOMDFileReader,
        'mode': 'rb'},
    'gtar': {
        'reader': reader.GetarFileReader,
        'mode': 'rb'},
    'dcd': {
        'reader': reader.DCDFileReader,
        'mode': 'rb'},
    'xml': {
        'reader': reader.HOOMDXMLFileReader,
        'mode': 'r'},
    'cif': {
        'reader': reader.CifFileReader,
        'mode': 'r'}}

# Mapping of format to writer classes and input modes
WRITE_CLASS_MODES = {
    'pos': {
        'writer': writer.PosFileWriter,
        'mode': 'w'},
    'gsd': {
        'writer': writer.GSDHOOMDFileWriter,
        'mode': 'wb'},
    'gtar': {
        'writer': writer.GetarFileWriter,
        'mode': 'w'},
    'cif': {
        'writer': writer.CifFileWriter,
        'mode': 'w'}}


def detect_format(filename):
    extension = os.path.splitext(filename)[1]
    try:
        file_format = FORMATS[extension]
    except KeyError:
        raise NotImplementedError(
            'The extension "{}" is not supported as a trajectory file.'.format(
                extension))
    return file_format


@contextmanager
def read(filename_or_fileobj, template=None, fmt=None):
    """This function reads a file and returns a trajectory, autodetecting the file format unless ``fmt`` is specified.

    :param filename_or_fileobj: Filename to read.
    :type filename_or_fileobj: string or file object
    :param template: Optional template for the GSDHOOMDFileReader.
    :type template: string
    :param fmt: File format, one of 'gsd', 'gtar', 'pos', 'cif', 'dcd', 'xml'
        (default: None, autodetected from filename_or_fileobj)
    :type fmt: string
    :returns: Trajectory read from the file.
    :rtype: :class:`~garnett.trajectory.Trajectory`
    """
    if isinstance(filename_or_fileobj, str):
        is_fileobj = False
        filename = filename_or_fileobj
    else:
        is_fileobj = True
        try:
            filename = filename_or_fileobj.name
        except AttributeError:
            raise RuntimeError(
                "Unable to determine filename from file object, "
                "which is required for format detection.")

    file_format = fmt or detect_format(filename)
    file_reader = READ_CLASS_MODES[file_format]['reader']()
    mode = READ_CLASS_MODES[file_format]['mode']

    with filename_or_fileobj if is_fileobj else open(filename_or_fileobj, mode) as read_file:
        if template is None:
            traj = file_reader.read(read_file)
            yield traj
        elif file_format == 'gsd':
            file_reader = READ_CLASS_MODES[file_format]['reader']()
            with read(template) as templatetraj:
                traj = file_reader.read(read_file, templatetraj[0])
                yield traj
        else:
            raise ValueError('The reader class {} does not support templates.'.format(file_reader.__class__.__name__))


def write(traj, filename_or_fileobj, fmt=None):
    """This function writes a trajectory to a file, autodetecting the file format unless ``fmt`` is specified.

    :param traj: Trajectory to write.
    :type traj: :class:`~garnett.trajectory.Trajectory`
    :param filename_or_fileobj: Filename to write.
    :type filename_or_fileobj: string or file object
    :param fmt: File format, one of 'gsd', 'gtar', 'pos', 'cif'
        (default: None, autodetected from filename_or_fileobj)
    :type fmt: string
    """
    if isinstance(filename_or_fileobj, str):
        is_fileobj = False
        filename = filename_or_fileobj
    else:
        is_fileobj = True
        try:
            filename = filename_or_fileobj.name
        except AttributeError:
            raise RuntimeError(
                "Unable to determine filename from file object, "
                "which is required for format detection.")

    file_format = detect_format(filename) if fmt is None else fmt
    file_writer = WRITE_CLASS_MODES[file_format]['writer']()
    mode = WRITE_CLASS_MODES[file_format]['mode']

    with filename_or_fileobj if is_fileobj else open(filename_or_fileobj, mode) as write_file:
        file_writer.write(traj, write_file)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from .posfilewriter import PosFileWriter
from .ciffilewriter import CifFileWriter

try:
    from .gsdhoomdfilewriter import GSDHOOMDFileWriter
except ImportError:
    class GSDHOOMDFileWriter(object):
        """Dummy implementation to provide useful errors for users"""
        def __init__(self):
            raise ImportError("Writing GSD files using garnett requires "
                              "the gsd package. Please install it if you "
                              "wish to write GSD files.")

try:
    from .getarfilewriter import GetarFileWriter
except ImportError:
    class GetarFileWriter(object):
        """Dummy implementation to provide useful errors for users"""
        def __init__(self):
            raise ImportError("Writing GTAR files using garnett requires "
                              "the libgetar package. Please install it if you "
                              "wish to write GTAR files.")

__all__ = ['PosFileWriter', 'CifFileWriter', 'GSDHOOMDFileWriter', 'GetarFileWriter']



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Hoomd-blue XML-file reader for the Glotzer Group, University of Michigan.

Authors: Carl Simon Adorf

.. code::

    reader = HOOMDXMLFileReader()
    with open('hoomdblue.xml') as xmlfile:
        return reader.read(xmlfile)
"""

import logging
import warnings
import xml.etree.ElementTree as ET

import numpy as np

from .trajectory import _RawFrameData, Frame, Trajectory, _generate_types_typeid
from .errors import ParserError


logger = logging.getLogger(__name__)


class HOOMDXMLFrame(Frame):

    def __init__(self, root):
        self.root = root
        super(HOOMDXMLFrame, self).__init__()

    def read(self):
        "Read the frame data from the stream."
        raw_frame = _RawFrameData()
        config = self.root.find('configuration')
        raw_frame.box = np.asarray(_get_box_matrix(config.find('box')))
        raw_frame.box_dimensions = int(config.get('dimensions', 3))
        raw_frame.position = list(_parse_position(config.find('position')))
        orientation = config.find('orientation')
        if orientation is not None:
            raw_frame.orientation = list(_parse_orientation(orientation))
        velocity = config.find('velocity')
        if velocity is not None:
            raw_frame.velocity = list(_parse_velocity(velocity))

        raw_frame.types, raw_frame.typeid = _parse_types(config.find('type'))
        return raw_frame

    def __str__(self):
        return "HOOMDXMLFrame(root={})".format(self.root)


class HOOMDXMLFileReader(object):
    "Reader for XML-files containing HOOMD-blue snapshots."

    def read(self, stream):
        """Read text stream and return a trajectory instance.

        :param stream: The stream, which contains the xmlfile.
        :type stream: A file-like textstream.
        """
        # Index the stream
        try:
            frames = [HOOMDXMLFrame(ET.fromstring(stream.read()))]
        except ET.ParseError as error:
            raise ParserError(error)
        logger.info("Read {} frames.".format(len(frames)))
        return Trajectory(frames)


def _get_box_matrix(box):
    assert box is not None

    def _get(name, default=None):
        v = box.get(name)
        if v is None and default is None:
            raise KeyError(name)
        elif v is None:
            return float(default)
        else:
            return float(v)
    return [
        [_get('lx'), _get('xy', 0) * _get('ly'), _get('xz', 0) * _get('lz')],
        [0.0, _get('ly'), _get('yz', 0) * _get('lz')],
        [0.0, 0.0, _get('lz')]
    ]


def _parse_position(position):
    for i, position_str in enumerate(position.text.splitlines()[1:]):
        yield [float(x) for x in position_str.split()]
    if i + 1 != int(position.attrib.get('num', i + 1)):
        warnings.warn("Number of positions is inconsistent.")


def _parse_velocity(velocity):
    for i, velocity_str in enumerate(velocity.text.splitlines()[1:]):
        yield [float(x) for x in velocity_str.split()]
    if i + 1 != int(velocity.attrib.get('num', i + 1)):
        warnings.warn("Number of velocities is inconsistent.")


def _parse_orientation(orientation):
    for i, orientation_str in enumerate(orientation.text.splitlines()[1:]):
        yield [float(x) for x in orientation_str.split()]
    if i + 1 != int(orientation.attrib.get('num', i + 1)):
        warnings.warn("Number of orientations is inconsistent.")


def _parse_types(types_element):
    type_strings = types_element.text.splitlines()[1:]
    types, typeid = _generate_types_typeid(type_strings)
    if len(typeid) != int(types_element.attrib.get('num', -1)):
        warnings.warn("Number of types is inconsistent.")
    return types, typeid



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""POS-file reader for the Glotzer Group, University of Michigan.

Authors: Carl Simon Adorf, Richmond Newmann

.. code::

    reader = PosFileReader()
    with open('a_posfile.pos', 'r', encoding='utf-8') as posfile:
        return reader.read(posfile)
"""

import collections
import logging
import warnings

import numpy as np

from .trajectory import _RawFrameData, Frame, Trajectory
from .shapes import FallbackShape, SphereShape, ArrowShape, SphereUnionShape, \
    PolygonShape, ConvexPolyhedronShape, ConvexSpheropolyhedronShape, \
    ConvexPolyhedronUnionShape, GeneralPolyhedronShape, EllipsoidShape
import rowan

from .errors import ParserError, ParserWarning

logger = logging.getLogger(__name__)

POSFILE_FLOAT_DIGITS = 11
COMMENT_CHARACTERS = ['//']
TOKENS_SKIP = ['translation', 'antiAliasing', 'zoomFactor', 'showEdges', 'connection']


def _is_comment(line):
    for comment_char in COMMENT_CHARACTERS:
        if line.startswith(comment_char):
            return True
    return False


class PosFileFrame(Frame):

    def __init__(self, stream, start, end, precision, default_type):
        self.stream = stream
        self.start = start
        self.end = end
        self.precision = precision
        self.default_type = default_type
        super(PosFileFrame, self).__init__()

    def _num(self, x):
        if isinstance(x, int):
            return x
        else:
            return round(float(x), self.precision)

    def _read_data_section(self, header, stream):
        """Read data section from stream."""
        data = collections.defaultdict(list)
        keys = header.strip().split()[1:]
        for i, line in enumerate(stream):
            if _is_comment(line):
                continue
            if line.startswith('#[done]'):
                return keys, data, i
            else:
                values = line.split()
                for key, value in zip(keys, values):
                    data[key].append(value)
        else:
            warnings.warn("File ended abruptly.", ParserWarning)
            return keys, data, i

    def _parse_shape_definition(self, line):
        tokens = (t for t in line.split())
        shape_class = next(tokens)
        if shape_class.lower() == 'sphere':
            diameter = float(next(tokens))
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            return SphereShape(diameter=diameter,
                               color=color)
        elif shape_class.lower() == 'arrow':
            thickness = float(next(tokens))
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            return ArrowShape(thickness=thickness,
                              color=color)
        elif shape_class.lower() == 'sphere_union':
            num_centers = int(next(tokens))
            centers = []
            diameters = []
            colors = []
            for i in range(num_centers):
                diameters.append(float(next(tokens)))
                xyz = next(tokens), next(tokens), next(tokens)
                colors.append(next(tokens))
                centers.append([self._num(v) for v in xyz])
            return SphereUnionShape(diameters=diameters,
                                    centers=centers,
                                    colors=colors)
        elif shape_class.lower() == 'poly3d_union':
            num_centers = int(next(tokens))
            vertices = [[] for p in range(num_centers)]
            centers = []
            orientations = []
            colors = []
            for i in range(num_centers):
                num_vertices = int(next(tokens))
                for j in range(num_vertices):
                    xyz = next(tokens), next(tokens), next(tokens)
                    vertices[i].append([self._num(v) for v in xyz])
                xyz = next(tokens), next(tokens), next(tokens)
                centers.append([self._num(v) for v in xyz])
                quat = next(tokens), next(tokens), next(tokens), next(tokens)
                orientations.append([self._num(q) for q in quat])
                colors.append(next(tokens))
            return ConvexPolyhedronUnionShape(vertices=vertices,
                                              centers=centers,
                                              orientations=orientations,
                                              colors=colors)
        elif shape_class.lower() == 'polyv':  # Officially polyV
            num_vertices = int(next(tokens))
            vertices = []
            for i in range(num_vertices):
                xyz = next(tokens), next(tokens), next(tokens)
                vertices.append([self._num(v) for v in xyz])
            num_faces = int(next(tokens))
            faces = []
            for i in range(num_faces):
                fv = []
                nvert = int(next(tokens))
                for j in range(nvert):
                    fv.append(int(next(tokens)))
                faces.append(fv)
            return GeneralPolyhedronShape(vertices=vertices,
                                          faces=faces)
        elif shape_class.lower() == 'poly3d':
            num_vertices = int(next(tokens))
            vertices = []
            for i in range(num_vertices):
                xyz = next(tokens), next(tokens), next(tokens)
                vertices.append([self._num(v) for v in xyz])
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            vertices = np.asarray(vertices)
            if (vertices[:, 2] == 0).all():
                # If the z-components of all vertices are zero,
                # create a 2D polygon instead
                return PolygonShape(vertices=vertices[:, :2],
                                    color=color)
            else:
                return ConvexPolyhedronShape(vertices=vertices,
                                             color=color)

            return ConvexPolyhedronShape(vertices=vertices,
                                         color=color)
        elif shape_class.lower() == 'spoly3d':
            rounding_radius = float(next(tokens))
            num_vertices = int(next(tokens))
            vertices = []
            for i in range(num_vertices):
                xyz = next(tokens), next(tokens), next(tokens)
                vertices.append([self._num(v) for v in xyz])
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            # Note: In POS files, there is no way to distinguish a 2D
            # spheropolygon with class spoly3d from a 3D spheropolyhedron whose
            # vertices lie in the x-y plane (with a rounding radius, it becomes
            # 3D).
            return ConvexSpheropolyhedronShape(vertices=vertices,
                                               rounding_radius=rounding_radius,
                                               color=color)
        elif shape_class.lower() == 'cyl':
            rounding_radius = float(next(tokens))/2
            height = float(next(tokens))
            if height > 0:
                vertices = [[-height/2, 0, 0], [height/2, 0, 0]]
            else:
                vertices = [[0, 0, 0]]
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            return ConvexSpheropolyhedronShape(vertices=vertices,
                                               rounding_radius=rounding_radius,
                                               color=color)
        elif shape_class.lower() == 'ellipsoid':
            a = float(next(tokens))
            b = float(next(tokens))
            c = float(next(tokens))
            try:
                color = next(tokens)
            except StopIteration:
                color = None
            return EllipsoidShape(a=a,
                                  b=b,
                                  c=c,
                                  color=color)
        else:
            warnings.warn("Failed to parse shape definition, "
                          "using fallback mode. ({})".format(line))
            return FallbackShape(line)

    def read(self):
        "Read the frame data from the stream."
        self.stream.seek(self.start)
        i = line = None

        def _assert(assertion):
            assert i is not None
            assert line is not None
            if not assertion:
                raise ParserError(
                    "Failed to read line #{}: {}.".format(i, line))
        monotype = False
        raw_frame = _RawFrameData()
        raw_frame.view_rotation = None
        for i, line in enumerate(self.stream):
            if _is_comment(line):
                continue
            if line.startswith('#'):
                if line.startswith('#[data]'):
                    _assert(raw_frame.data is None)
                    raw_frame.data_keys, raw_frame.data, j = \
                        self._read_data_section(line, self.stream)
                    i += j
                else:
                    raise ParserError(line)
            else:
                tokens = line.rstrip().split()
                if not tokens:
                    continue  # empty line
                elif tokens[0] in TOKENS_SKIP:
                    continue  # skip these lines
                if tokens[0] == 'eof':
                    logger.debug("Reached end of frame.")
                    break
                elif tokens[0] == 'def':
                    definition, data, end = line.strip().split('"')
                    _assert(len(end) == 0)
                    name = definition.split()[1]
                    type_shape = self._parse_shape_definition(data)
                    if name not in raw_frame.types:
                        raw_frame.types.append(name)
                        raw_frame.type_shapes.append(type_shape)
                    else:
                        typeid = raw_frame.type_shapes.index(name)
                        raw_frame.type_shapes[typeid] = type_shape
                        warnings.warn("Redefinition of type '{}'.".format(name))
                elif tokens[0] == 'shape':  # monotype
                    data = line.strip().split('"')[1]
                    raw_frame.types.append(self.default_type)
                    type_shape = self._parse_shape_definition(data)
                    raw_frame.type_shapes.append(type_shape)
                    _assert(len(raw_frame.type_shapes) == 1)
                    monotype = True
                elif tokens[0] in ('boxMatrix', 'box'):
                    if len(tokens) == 10:
                        raw_frame.box = np.array(
                            [self._num(v) for v in tokens[1:]]).reshape((3, 3))
                    elif len(tokens) == 4:
                        raw_frame.box = np.array([
                            [self._num(tokens[1]), 0, 0],
                            [0, self._num(tokens[2]), 0],
                            [0, 0, self._num(tokens[3])]]).reshape((3, 3))
                elif tokens[0] == 'rotation':
                    euler_angles = np.array([float(t) for t in tokens[1:]])
                    euler_angles *= np.pi / 180
                    raw_frame.view_rotation = rowan.from_euler(*euler_angles, axis_type='extrinsic', convention='xyz')
                else:
                    # assume we are reading positions now
                    if not monotype:
                        name = tokens[0]
                        if name not in raw_frame.types:
                            raw_frame.types.append(name)
                            type_shape = self._parse_shape_definition(' '.join(tokens[:3]))
                            raw_frame.type_shapes.append(type_shape)
                    else:
                        name = self.default_type
                    typeid = raw_frame.types.index(name)
                    type_shape = raw_frame.type_shapes[typeid]
                    if len(tokens) == 7 and isinstance(type_shape, ArrowShape):
                        xyz = tokens[-6:-3]
                        quat = tokens[-3:] + [0]
                    elif len(tokens) >= 7:
                        xyz = tokens[-7:-4]
                        quat = tokens[-4:]
                    elif len(tokens) >= 3:
                        xyz = tokens[-3:]
                        quat = None
                    else:
                        raise ParserError(line)
                    raw_frame.typeid.append(typeid)
                    raw_frame.position.append([self._num(v) for v in xyz])
                    if quat is None:
                        raw_frame.orientation.append(quat)
                    else:
                        raw_frame.orientation.append([self._num(v) for v in quat])

        # Perform inverse rotation to recover original coordinates
        if raw_frame.view_rotation is not None:
            pos = rowan.rotate(rowan.inverse(raw_frame.view_rotation), raw_frame.position)
        else:
            pos = np.asarray(raw_frame.position)
        # If all the z coordinates are close to zero, set box dimension to 2
        if np.allclose(pos[:, 2], 0.0, atol=1e-7):
            raw_frame.box_dimensions = 2

        # If no valid orientations have been added, the array should be empty
        if all([quat is None for quat in raw_frame.orientation]):
            raw_frame.orientation = []
        else:
            # Replace values of None with an identity quaternion
            for i in range(len(raw_frame.orientation)):
                if raw_frame.orientation[i] is None:
                    raw_frame.orientation[i] = [1, 0, 0, 0]
        return raw_frame

    def __str__(self):
        return "PosFileFrame(stream={}, start={}, end={})".format(
            self.stream, self.start, self.end)


class PosFileReader(object):
    """POS-file reader for the Glotzer Group, University of Michigan.

        Authors: Carl Simon Adorf, Richmond Newmann

        .. code::

            reader = PosFileReader()
            with open('a_posfile.pos', 'r', encoding='utf-8') as posfile:
                return reader.read(posfile)

        :param precision: The number of digits to
                          round floating-point values to.
        :type precision: int"""

    def __init__(self, precision=None):
        """Initialize a pos-file reader.

        :param precision: The number of digits to
                          round floating-point values to.
        :type precision: int
        """
        self._precision = precision or POSFILE_FLOAT_DIGITS

    def _scan(self, stream, default_type):
        start = 0
        index = 0
        for line in stream:
            index += len(line)
            if line.startswith('eof'):
                yield PosFileFrame(
                    stream, start, index,
                    self._precision, default_type)
                start = index
        if index > start:
            stream.seek(start)
            for line in stream:
                if line.startswith('boxMatrix') or line.startswith('box'):
                    stream.seek(0, 2)
                    yield PosFileFrame(
                            stream, start, index,
                            self._precision, default_type)
                    break
            else:
                logger.warning("Unexpected file ending.")

    def read(self, stream, default_type='A'):
        """Read text stream and return a trajectory instance.

        :param stream: The stream, which contains the posfile.
        :type stream: A file-like textstream.
        :param default_type: The default particle type for
                             posfile dialects without type definition.
        :type default_type: str
        """
        # Index the stream
        frames = list(self._scan(stream, default_type))
        if len(frames) == 0:
            raise ParserError("Did not read a single complete frame.")
        logger.info("Read {} frames.".format(len(frames)))
        return Trajectory(frames)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""DCD-file reader for the Glotzer Group, University of Michigan.

Authors: Carl Simon Adorf

A dcd file consists only of positions.
To provide additional information it is possible
to provide a frame object, whose properties
are copied into each frame of the dcd trajectory.

The example is given for a hoomd-blue xml frame:

.. code::

    xml_reader = HOOMDXMLFileReader()
    dcd_reader = DCDFileReader()

    with open('init.xml') as xmlfile:
        with open('dump.dcd', 'rb') as dcdfile:
            xml_frame = xml_reader.read(xmlfile)[0]
            traj = reader.read(dcdfile, xml_frame)

.. note::

    If the topology frame is 2-dimensional, the dcd
    trajectory positions are interpreted such that
    the first two values contain the xy-coordinates,
    the third value is an euler angle.

    The euler angle is converted to a quaternion and stored
    in the orientation of the frame.

    To retrieve the euler angles, simply convert the quaternion:

    .. code::

        ort = traj[0].orientation
        alpha = 2 * np.arctan2(ort[:, 3], ort[:, 0])
"""

import logging
import copy
from collections import namedtuple

import numpy as np
from numpy.core import numeric as _nx
from numpy.core.numeric import asanyarray

from .trajectory import Frame, Trajectory, Box
from .trajectory import FRAME_PROPERTIES, TYPE_PROPERTIES, PARTICLE_PROPERTIES
from .trajectory import _RawFrameData
from . import pydcdreader

logger = logging.getLogger(__name__)


def _euler_to_quaternion(alpha, q):
    q.T[0] = np.cos(alpha * 0.5)
    q.T[1] = q.T[2] = 0.0
    q.T[3] = np.sin(alpha * 0.5)


def _box_from_frame_header(frame_header):
    fh = frame_header

    lx = fh.box_a
    xy = fh.box_b * fh.box_gamma
    xz = fh.box_c * fh.box_beta
    ly = np.sqrt(fh.box_b*fh.box_b - xy*xy)
    yz = (fh.box_b*fh.box_c*fh.box_alpha - xy*xz) / lx
    lz = np.sqrt(fh.box_c*fh.box_c - xz*xz - yz*yz)

    xy /= ly
    xz /= lz
    yz /= lz
    return Box(lx, ly, lz, xy=xy, xz=xz, yz=yz, dimensions=3)


def _np_stack(arrays, axis=0):
    """Join a sequence of arrays along a new axis.

    Copyright (c) 2005-2016, NumPy Developers.
    All rights reserved.

    From numpy/core/shape_base.py (requires numpy>=1.10),
    included to avoid numpy dependency >=1.10."""
    arrays = [asanyarray(arr) for arr in arrays]
    if not arrays:
        raise ValueError('need at least one array to stack')

    shapes = set(arr.shape for arr in arrays)
    if len(shapes) != 1:
        raise ValueError('all input arrays must have the same shape')

    result_ndim = arrays[0].ndim + 1
    if not -result_ndim <= axis < result_ndim:
        msg = 'axis {0} out of bounds [-{1}, {1})'.format(axis, result_ndim)
        raise IndexError(msg)
    if axis < 0:
        axis += result_ndim

    sl = (slice(None),) * axis + (_nx.newaxis,)
    expanded_arrays = [arr[sl] for arr in arrays]
    return _nx.concatenate(expanded_arrays, axis=axis)


_DCDFileHeader = namedtuple(
    '_DCDFileHeader',
    ('num_frames', 'm_start_timestep', 'm_period', 'timesteps',
     'timestep', 'include_unitcell', 'charmm_version', 'n_particles'))


_DCDFrameHeader = namedtuple(
    '_DCDFrameHeader',
    ('box_a', 'box_gamma', 'box_b', 'box_beta', 'box_alpha', 'box_c'))


class DCDFrame(Frame):

    def __init__(self, dcdreader, stream, file_header,
                 offset, t_frame, default_type='A',
                 dtype=None):
        self._dcdreader = dcdreader
        self.stream = stream
        self.file_header = _DCDFileHeader(** file_header)
        self.offset = offset
        self.t_frame = t_frame
        self.default_type = default_type
        self._box = None
        self._types = None
        self._typeid = None
        self._position = None
        self._orientation = None
        super(DCDFrame, self).__init__(dtype=dtype)

    def __len__(self):
        return int(self.file_header.n_particles)

    def _read(self, xyz):
        frame_header = _DCDFrameHeader(
            ** self._dcdreader.read_frame(self.stream, xyz, self.offset))
        self._box = _box_from_frame_header(frame_header)
        self._position = xyz.swapaxes(0, 1)

    def _load(self, xyz=None, ort=None):
        N = FRAME_PROPERTIES['N'](self.file_header.n_particles)
        if xyz is None:
            xyz = np.zeros((3, N), dtype=np.float32)
        if ort is None:
            ort = np.zeros((N, 4), dtype=self._dtype)
        self._read(xyz=xyz)
        if self.t_frame is None:
            self._types = np.asarray([self.default_type],
                                     dtype=TYPE_PROPERTIES['types'])
            self._typeid = np.zeros(N, dtype=PARTICLE_PROPERTIES['typeid'])
        else:
            self._types = copy.copy(self.t_frame.types)
            self._typeid = copy.copy(self.t_frame.typeid)
        if self.t_frame is None or self.t_frame.box.dimensions == 3:
            ort.T[0] = 1.0
            ort.T[1:] = 0
        elif self.t_frame.box.dimensions == 2:
            _euler_to_quaternion(
                self._position.T[-1], ort)
            self._position.T[-1] = 0
        else:
            raise ValueError(self.t_frame.box.dimensions)
        self._orientation = ort

    def _loaded(self):
        return not (self._box is None or
                    self._types is None or
                    self._typeid is None or
                    self._position is None or
                    self._orientation is None)

    def read(self):
        raw_frame = _RawFrameData()
        if self.t_frame is not None:
            raw_frame.data = copy.deepcopy(self.t_frame.data)
            raw_frame.data_keys = copy.deepcopy(self.t_frame.data_keys)
            raw_frame.box_dimensions = self.t_frame.box.dimensions
            try:
                raw_frame.type_shapes = copy.deepcopy(self.t_frame.type_shapes)
            except AttributeError:
                pass
        if not self._loaded():
            self._load()
        assert self._loaded()
        raw_frame.box = self._box
        raw_frame.types = self._types
        raw_frame.typeid = self._typeid
        raw_frame.position = self._position
        raw_frame.orientation = self._orientation
        assert len(raw_frame.typeid) == len(self)
        assert len(raw_frame.position) == len(self)
        assert len(raw_frame.orientation) == len(self)
        return raw_frame

    def unload(self):
        self._box = None
        self._types = None
        self._position = None
        self._orientation = None
        super(DCDFrame, self).unload()

    def __str__(self):
        return "DCDFrame(# particles={}, topology_frame={})".format(
            len(self), self.t_frame)


class DCDTrajectory(Trajectory):

    def arrays_loaded(self):
        """Returns true if arrays are loaded into memory.

        See also: :meth:`~.load_arrays`"""
        return not (self._box is None or
                    self._N is None or
                    self._types is None or
                    self._typeid is None or
                    self._position is None or
                    self._orientation is None)

    def load_arrays(self):
        # Determine array shapes
        M = len(self)
        N = len(self.frames[0])
        _N = np.full(M, N, dtype=FRAME_PROPERTIES['N'])

        # Coordinates
        xyz = np.zeros((M, 3, N), dtype=np.float32)
        ort = np.zeros((M, N, 4), dtype=self._dtype)
        for i, frame in enumerate(self.frames):
            if not frame._loaded():
                frame._load(xyz=xyz[i], ort=ort[i])

        # Types can only be handled after frame._load() calls.
        types = np.asarray([f._types for f in self.frames],
                           dtype=TYPE_PROPERTIES['types'])
        typeid = np.asarray([f._typeid for f in self.frames],
                            dtype=PARTICLE_PROPERTIES['typeid'])
        box = np.asarray([f._box for f in self.frames],
                         dtype=FRAME_PROPERTIES['box'])

        try:
            # Perform swap
            self._box = box
            self._N = _N
            self._types = types
            self._typeid = typeid
            self._position = xyz.swapaxes(1, 2)
            self._orientation = ort
        except Exception:
            # Ensure consistent error state
            self._box = self._N = self._type = self._types = \
                self._type_ids = self._position = self._orientation = None
            raise

    def xyz(self, xyz=None):
        """Return the xyz coordinates of the dcd file.

        Use this function to access xyz-coordinates with minimal
        overhead and maximal performance.

        You can provide a reference to an existing numpy.ndarray
        with shape (Mx3xN), where M is the length of the trajectory
        and N is the number of particles.
        Please note that the array needs to be of data type float32
        and in-memory contiguous.

        :param xyz: A numpy array of shape (Mx3xN).
        :type xyz: numpy.ndarray
        :returns: A view or a copy of the xyz-array of shape (MxNx3).
        :rtype: numpy.ndarray
        """
        shape = (len(self), 3, len(self.frames[0]))
        if xyz is None:
            xyz = np.zeros(shape, dtype=np.float32)
        else:
            assert xyz.flags['C_CONTIGUOUS']
            assert xyz.dtype == np.float32
            assert xyz.shape == shape
        for i, frame in enumerate(self.frames):
            frame._read(xyz[i])
        return xyz.swapaxes(1, 2)


class _DCDFileReader(object):
    """DCD-file reader for the Glotzer Group, University of Michigan.

    Author: Carl Simon Adorf

    A dcd file consists only of positions.
    To provide additional information it is possible
    to provide a frame object, whose properties
    are copied into each frame of the dcd trajectory.

    The example is given for a hoomd-blue xml frame:

    .. code::

        xml_reader = HOOMDXMLFileReader()
        dcd_reader = DCDFileReader()

        with open('init.xml') as xmlfile:
            with open('dump.dcd', 'rb') as dcdfile:
                xml_frame = xml_reader.read(xmlfile)[0]
                traj = reader.read(dcdfile, xml_frame)

    .. note::

        If the topology frame is 2-dimensional, the dcd
        trajectory positions are interpreted such that
        the first two values contain the xy-coordinates,
        the third value is an euler angle.

        The euler angle is converted to a quaternion and stored
        in the orientation of the frame.

        To retrieve the euler angles, simply convert the quaternion:

        .. code::

            alpha = 2 * np.arctan2(ort[:, 3], ort[:, 0])
    """
    _dcdreader = pydcdreader

    def _scan(self, stream, t_frame=None, default_type=None):
        file_header, offsets = self._dcdreader.scan(stream)
        for offset in offsets:
            yield DCDFrame(
                dcdreader=self._dcdreader,
                stream=stream, file_header=file_header, offset=offset,
                t_frame=t_frame, default_type=default_type)

    def read(self, stream, frame=None, default_type=None):
        """Read binary stream and return a trajectory instance.

        :param stream: The stream, which contains the dcd-file.
        :type stream: A file-like binary stream
        :param frame: A frame containing topological information
            that cannot be encoded in the dcd-format.
        :type frame: :class:`trajectory.Frame`
        :param default_type: A type name to be used when no first
            frame is provided, defaults to 'A'.
        :type default_type: str
        :returns: A trajectory instance.
        :rtype: :class:`~.DCDTrajectory`"""
        if frame is not None and default_type is not None:
            raise ValueError(
                "You can only provide a frame or a default_type, not both.")
        elif frame is None and default_type is None:
            default_type = 'A'
        elif frame is not None and frame.box.dimensions == 2:
            logger.info(
                "2-dimensional box, interpreting 3rd dimension "
                "as euler orientation angle.")

        frames = list(self._scan(stream, t_frame=frame,
                                 default_type=default_type))
        logger.info("Read {} frames.".format(len(frames)))
        return DCDTrajectory(frames)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""getar-file reader for the Glotzer Group, University of Michigan.

Authors: Matthew Spellings, Carl Simon Adorf

.. code::

    reader = GetarFileReader()
    traj = reader.read(open('trajectory.tar', 'rb'))
"""

import bisect
import functools
import gtar
import json
import logging
import numpy as np

from .trajectory import _RawFrameData, Box, Frame, Trajectory
from .shapes import _parse_type_shape

logger = logging.getLogger(__name__)


def _find_le(a, x):
    """Find rightmost value less than or equal to x"""
    i = bisect.bisect_right(a, x)
    if i:
        return a[i-1]
    raise ValueError


@functools.lru_cache(maxsize=64)
def _get_record_frames(trajectory, record):
    return trajectory.queryFrames(record)


class GetarFrame(Frame):
    """Interface to grab getar frame data.

    :param trajectory: gtar.GTAR trajectory object to read from
    :param records: Dictionary of gtar.Record objects
    :param frame: Frame name inside the trajectory
    :param default_type: The default particle type
    :type default_type: str
    """

    def __init__(self, trajectory, records, frame, default_type, default_box):
        super(GetarFrame, self).__init__()
        self._trajectory = trajectory
        self._records = records
        self._frame = frame
        self._default_type = default_type
        self._default_box = default_box

    def __str__(self):
        return "GetarFrame({})".format(self._records)

    def _get_record_frame(self, name):
        record = self._records[name]
        frames = _get_record_frames(self._trajectory, record)
        return _find_le(frames, self._frame)

    def _get_record_value(self, name):
        record = self._records[name]
        frame = self._get_record_frame(name)
        return self._trajectory.getRecord(record, frame)

    def read(self):
        raw_frame = _RawFrameData()
        raw_frame.type_shapes = []

        # Map getar field names onto HOOMD convention's names
        prop_map = {
            'angular_momentum_quat': 'angmom',
            'type': 'typeid',
        }

        supported_records = [
            'type', 'position', 'orientation', 'velocity',
            'mass', 'charge', 'diameter',
            'moment_inertia', 'angular_momentum_quat',
            'image'
        ]
        for name in supported_records:
            try:
                values = self._get_record_value(name)
                frame_prop = prop_map.get(name, name)
                setattr(raw_frame, frame_prop, values)
            except (KeyError, ValueError):
                # either the record wasn't present (KeyError) or the
                # quantity was not yet provided in index-order
                # (ValueError)
                pass

        try:
            raw_frame.types = json.loads(self._get_record_value('type_names.json'))
        except (KeyError, ValueError):
            raw_frame.types = [self._default_type]
        # if raw_frame.typeid is not set, np.max() here returns a
        # float, so pass through int()
        num_types_to_add = (int(np.max(raw_frame.typeid, initial=0)) +
                            1 - len(raw_frame.types))
        last_type_name = raw_frame.types[-1]
        for t in range(1, num_types_to_add + 1):
            type_name = (last_type_name[:-1] + chr(ord(last_type_name[-1]) + t))
            raw_frame.types.append(type_name)

        try:
            # Read dimension if stored
            if 'dimensions' in self._records:
                dimensions = self._get_record_value('dimensions')[0]
            # Fallback to detection based on z coordinates
            else:
                zs = raw_frame.position[:, 2]
                dimensions = 2 if np.allclose(zs, 0.0, atol=1e-7) else 3

            box = self._get_record_value('box')
            gbox = Box(
                **dict(
                    zip(['Lx', 'Ly', 'Lz', 'xy', 'xz', 'yz'], box),
                    dimensions=dimensions)
            )
            raw_frame.box = np.array(gbox.get_box_matrix())
            raw_frame.box_dimensions = int(dimensions)
        except (KeyError, ValueError):
            raw_frame.box = self._default_box

        try:
            type_shapes = json.loads(self._get_record_value('type_shapes.json'))
            raw_frame.type_shapes = [_parse_type_shape(t) for t in type_shapes]
        except (KeyError, ValueError):
            pass

        return raw_frame


class GetarFileReader(object):
    """getar-file reader for the Glotzer Group, University of Michigan.

    Authors: Matthew Spellings, Carl Simon Adorf

    Read GEneric Trajectory ARchive files, a binary format designed
    for efficient, extensible storage of trajectory data.

    This class provides a wrapper for the gtar library.

    .. code::

        reader = GetarFileReader()
        with open('trajectory.tar', 'rb') as file:
            traj = reader.read(file)
    """

    def read(self, stream, default_type='A', default_box=None):
        """Read binary stream and return a trajectory instance.

        :param stream: The stream, which contains the GeTarFile.
        :type stream: A file-like binary stream.
        :param default_type: The default particle type for
                             posfile dialects without type definition.
        :type default_type: str
        :param default_box: The default_box value is used if no
                            box is specified in the libgetar file.
                            Defaults to [Lx=Ly=Lz=1.0].
        :type default_box: :class:`numpy.ndarray`
        """
        if default_box is None:
            default_box = np.diag([1.0] * 3)
        try:
            filename = stream.name
        except AttributeError:
            raise NotImplementedError(
                "The current implementation of the GeTarFileReader requires "
                "file objects with name attribute, such as NamedTemporaryFile "
                "as the underlying library is reading the file by filename "
                "and not directly from the stream.")
        _trajectory = gtar.GTAR(filename, 'r')
        _records = {rec.getName(): rec for rec in _trajectory.getRecordTypes() if not rec.getGroup()}
        # assume that we care primarily about positions
        try:
            self._frames = _trajectory.queryFrames(_records['position'])
        except KeyError:
            raise RuntimeError("Given trajectory '{}' contained no "
                               "positions.".format(stream))
        frames = [GetarFrame(_trajectory, _records, idx, default_type, default_box)
                  for idx in self._frames]
        logger.info("Read {} frames.".format(len(frames)))
        return Trajectory(frames)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""getar-file writer for the Glotzer Group, University of Michigan.

Authors: Bradley Dice

.. code::

    writer = GetarFileWriter()
    with open('trajectory.tar', 'wb') as gtarfile:
        writer.write(trajectory, gtarfile)
"""

from gtar import GTAR, Record
import json
import logging
logger = logging.getLogger(__name__)


class GetarFileWriter(object):
    """getar-file writer for the Glotzer Group, University of Michigan.

    Authors: Bradley Dice

    Write GEneric Trajectory ARchive files, a binary format designed
    for efficient, extensible storage of trajectory data.

    This class provides a wrapper for the gtar library.

    .. code::

        writer = GetarFileWriter()
        with open('trajectory.tar', 'wb') as gtarfile:
            writer.write(trajectory, gtarfile)

        # For appending to the file
        with open('trajectory.tar', 'ab') as gtarfile:
            writer.write(trajectory, gtarfile)
    """

    property_record_map = {
        'typeid': 'type.u32.ind',
        'position': 'position.f32.ind',
        'orientation': 'orientation.f32.ind',
        'velocity': 'velocity.f32.ind',
        'mass': 'mass.f32.ind',
        'charge': 'charge.f32.ind',
        'diameter': 'diameter.f32.ind',
        'moment_inertia': 'moment_inertia.f32.ind',
        'angmom': 'angular_momentum_quat.f32.ind',
        'image': 'image.i32.ind',
        }

    def makeRecord(self, name, index=None, prefix=None):
        if index is not None:
            frame = 'frames/{}'.format(index)
        else:
            frame = None
        path_parts = [prefix, frame, name]
        path = '/'.join(filter(lambda x: x is not None, path_parts))
        return Record(path)

    def writeFrame(self, bulkwriter, frame, index=None, skip_props=False):
        """Write the frame data for an index using a bulk writer."""

        # Type names
        types_rec = self.makeRecord('type_names.json', index=index)
        types_contents = json.dumps(frame.types)
        bulkwriter.writeRecord(rec=types_rec, contents=types_contents)

        if not skip_props:
            # Particle properties
            for prop, recname in type(self).property_record_map.items():
                rec = self.makeRecord(recname, index=index)
                contents = getattr(frame, prop)
                bulkwriter.writeRecord(rec=rec, contents=contents)

        # Box and dimensions
        box_rec = self.makeRecord('box.f32.uni', index=index)
        bulkwriter.writeRecord(rec=box_rec, contents=frame.box.get_box_array())
        dim_rec = self.makeRecord('dimensions.u32.uni', index=index)
        bulkwriter.writeRecord(rec=dim_rec, contents=[frame.box.dimensions])

        # Shape definitions
        type_shapes = getattr(frame, 'type_shapes', None)
        if type_shapes is not None:
            shape_rec = self.makeRecord('type_shapes.json', index=index)
            type_shapes = json.dumps([t.type_shape for t in type_shapes])
            bulkwriter.writeRecord(rec=shape_rec, contents=type_shapes)

    def write(self, trajectory, stream, static_frame=None):
        """Serialize a trajectory into gtar-format and write it to a file.

        :param trajectory: The trajectory to serialize
        :type trajectory: :class:`~garnett.trajectory.Trajectory`
        :param filename: The file to write to."""

        try:
            filename = stream.name
            mode = stream.mode
        except AttributeError:
            raise NotImplementedError(
                "The current implementation of the GetarFileWriter requires "
                "file objects with name attribute, such as NamedTemporaryFile "
                "as the underlying library is reading the file by filename "
                "and not directly from the stream.")
        with GTAR(path=filename, mode=mode) as t, \
                t.getBulkWriter() as t_writer:

            for index, frame in enumerate(trajectory):

                if index == 0:  # avoid indexing to allow tqdm(trajectory)
                    if static_frame is None:
                        # Write the first frame of the trajectory as static data
                        # so that box, type, and shape information is accessible
                        self.writeFrame(t_writer, frame,
                                        index=None, skip_props=True)
                    else:
                        self.writeFrame(t_writer, static_frame,
                                        index=None, skip_props=True)

                self.writeFrame(t_writer, frame, index)

                logger.debug("Wrote frame {}.".format(index + 1))

        logger.info("Wrote {} frames.".format(index + 1))



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""
This is a collection of samples, parsers and writers for formats
used in the Glotzer Group at the University of Michigan, Ann Arbor."""


from . import reader
from . import writer
from . import samples
from . import shapes
from . import trajectory
from .util import read, write
from .version import __version__


__all__ = [
    '__version__',
    'reader',
    'writer',
    'samples',
    'shapes',
    'trajectory',
    'read',
    'write',
]



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import struct


class _DCDFileHeader(object):
    pass


class _DCDFrameHeader(object):
    pass


def _read_int(stream):
    return struct.unpack('<L', stream.read(4))[0]


def _read_double(stream):
    return struct.unpack('<d', stream.read(8))[0]


def _read_float(stream):
    return struct.unpack('<f', stream.read(4))[0]


def _read_file_header(stream):
    file_header = _DCDFileHeader()
    assert _read_int(stream) == 84
    assert stream.read(4) == b'CORD'
    file_header.num_frames = _read_int(stream)
    file_header.m_start_timestep = _read_int(stream)
    file_header.m_period = _read_int(stream)
    file_header.timesteps = _read_int(stream)
    for i in range(5):
        _read_int(stream)
    file_header.timestep = _read_int(stream)
    file_header.include_unitcell = bool(_read_int(stream))
    for i in range(8):
        assert _read_int(stream) == 0
    file_header.charmm_version = _read_int(stream)
    assert _read_int(stream) == 84
    title_section_size = _read_int(stream)
    n_titles = _read_int(stream)
    len_title = int((title_section_size - 4) / 2)
    for i in range(n_titles):
        stream.read(len_title)
    assert _read_int(stream) == title_section_size
    assert _read_int(stream) == 4
    file_header.n_particles = _read_int(stream)
    assert _read_int(stream) == 4
    return file_header


def _skip_frame(stream):
    for i in range(3):
        len_section = _read_int(stream)
        stream.seek(len_section, 1)
        assert _read_int(stream) == len_section


def _read_frame_header(stream):
    frame_header = _DCDFrameHeader()
    frame_header_size = _read_int(stream)
    frame_header.box_a = _read_double(stream)
    frame_header.box_gamma = _read_double(stream)
    frame_header.box_b = _read_double(stream)
    frame_header.box_beta = _read_double(stream)
    frame_header.box_alpha = _read_double(stream)
    frame_header.box_c = _read_double(stream)
    assert _read_int(stream) == frame_header_size
    return frame_header


def _read_frame_body(stream, xyz):
    N = xyz.shape[1]
    for i in range(3):
        len_section = _read_int(stream)
        for j in range(N):
            xyz[i][j] = _read_float(stream)
        assert _read_int(stream) == len_section


def scan(stream):
    file_header = _read_file_header(stream)
    offsets = []
    for i in range(file_header.num_frames):
        offsets.append(stream.tell())
        _read_frame_header(stream)
        _skip_frame(stream)
    return file_header.__dict__, offsets


def read_frame(stream, xyz, offset=-1):
    if offset >= 0:
        stream.seek(offset)
    frame_header = _read_frame_header(stream)
    _read_frame_body(stream, xyz)
    return frame_header.__dict__



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

"""Abstract shape definitions used to read/write particle shapes."""

import json
import logging
import numpy as np

__all__ = [
    'FallbackShape',
    'Shape',
    'SphereShape',
    'ArrowShape',
    'SphereUnionShape',
    'PolygonShape',
    'SpheropolygonShape',
    'ConvexPolyhedronShape',
    'ConvexPolyhedronUnionShape',
    'ConvexSpheropolyhedronShape',
    'GeneralPolyhedronShape',
    'EllipsoidShape',
]

logger = logging.getLogger(__name__)

SHAPE_DEFAULT_COLOR = '005984FF'


class _NumpyEncoder(json.JSONEncoder):
    """JSONEncoder class converting NumPy arrays to lists."""
    def default(self, obj):
        if isinstance(obj, np.number):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def _json_sanitize(func):
    """Decorator ensuring that returned data is JSON-encodable."""
    def wrapper(*args, **kwargs):
        data = func(*args, **kwargs)
        return json.loads(json.dumps(data, cls=_NumpyEncoder))
    # Ensure that the decorated function inherits the intended docstring.
    wrapper.__doc__ = func.__doc__
    return wrapper


class FallbackShape(str):
    """This shape definition class is used when no specialized Shape class can be applied.

    The fallback shape definition is a string containing the definition."""
    pass


class Shape(object):
    """Parent class of all shape objects.

    :param shape_class:
        Shape class directive, used for POS format (default: :code:`None`).
    :type shape_class:
        str
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, shape_class=None, color=None):
        self.shape_class = shape_class
        self.color = color if color else SHAPE_DEFAULT_COLOR

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError as e:
            raise KeyError(*e.args)

    @property
    def pos_string(self):
        return "{} {}".format(self.shape_class, self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        return {"type": self.shape_class}

    def __str__(self):
        return json.dumps(self.type_shape)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.type_shape == other.type_shape


class SphereShape(Shape):
    """Shape class for spheres of a specified diameter.

    :param diameter:
        Diameter of the sphere.
    :type diameter:
        float
    :param orientable:
         Set to True for spheres with orientation (default: :code:`False`).
    :type orientable:
        bool
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, diameter, orientable=False, color=None):
        super(SphereShape, self).__init__(
            shape_class='sphere', color=color)
        self.diameter = diameter
        self.orientable = orientable

    @property
    def pos_string(self):
        return "{} {} {}".format(self.shape_class, self.diameter, self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> SphereShape(2.0).type_shape
            {'type': 'Sphere', 'diameter': 2.0}
        """
        return {'type': 'Sphere',
                'diameter': self.diameter}


class ArrowShape(Shape):
    """Shape class for arrows of a specified thickness.

    :param thickness:
        Thickness of the arrow.
    :type thickness:
        float
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, thickness=0.1, color=None):
        super(ArrowShape, self).__init__(
            shape_class='arrow', color=color)
        self.thickness = thickness

    @property
    def pos_string(self):
        return "{} {} {}".format(self.shape_class, self.thickness, self.color)


class SphereUnionShape(Shape):
    """Shape class for sphere unions, such as rigid bodies of many spheres.

    :param diameters:
        List of sphere diameters.
    :type diameters:
        list
    :param centers:
        List of 3D center vectors.
    :type centers:
        list
    :param colors:
        List of hexadecimal color strings in format :code:`RRGGBBAA` (default: :code:`None`).
    :type colors:
        list
    """

    def __init__(self, diameters, centers, colors=None):
        super(SphereUnionShape, self).__init__(
            shape_class='sphere_union', color='')
        self.diameters = diameters
        self.centers = centers
        self.colors = colors

    @property
    def pos_string(self):
        shape_def = '{} {} '.format(self.shape_class, len(self.centers))
        for d, p, c in zip(self.diameters, self.centers, self.colors):
            shape_def += '{0} '.format(d)
            shape_def += '{0} {1} {2} '.format(*p)
            shape_def += '{0} '.format(c)

        return shape_def

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

           >>> SphereUnionShape([0.5, 0.5, 0.5], [[0, 0, 1.0], [0, 1.0, 0], [1.0, 0, 0]]).type_shape
           {'type': 'SphereUnion', 'diameters': [0.5, 0.5, 0.5],
            'centers': [[0, 0, 1.0], [0, 1.0, 0], [1.0, 0, 0]]}
        """
        return {'type': 'SphereUnion',
                'diameters': self.diameters,
                'centers': self.centers}


class PolygonShape(Shape):
    """Shape class for polygons in a 2D plane.

    :param vertices:
        List of 2D vertex vectors.
    :type vertices:
        list
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, vertices, color=None):
        super(PolygonShape, self).__init__(
            shape_class='poly3d', color=color)
        self.vertices = vertices

    @property
    def pos_string(self):
        return "{} {} {} {}".format(
            self.shape_class,
            len(self.vertices),
            ' '.join('{} {} 0'.format(v[0], v[1]) for v in self.vertices),
            self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> PolygonShape([[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5]]).type_shape
            {'type': 'Polygon', 'rounding_radius': 0,
             'vertices': [[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5]]}
        """
        return {'type': 'Polygon',
                'rounding_radius': 0,
                'vertices': self.vertices}


class SpheropolygonShape(Shape):
    """Shape class for rounded polygons in a 2D plane.

    :param vertices:
        List of 2D vertex vectors.
    :type vertices:
        list
    :param rounding_radius:
        Rounding radius applied to the spheropolygon (default: 0).
    :type rounding_radius:
        float
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, vertices, rounding_radius=0, color=None):
        super(SpheropolygonShape, self).__init__(
            shape_class='spoly3d', color=color)
        self.vertices = vertices
        self.rounding_radius = rounding_radius

    @property
    def pos_string(self):
        return "{} {} {} {} {}".format(
            self.shape_class,
            self.rounding_radius,
            len(self.vertices),
            ' '.join('{} {} 0'.format(v[0], v[1]) for v in self.vertices),
            self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> SpheropolygonShape([[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5]], 0.1).type_shape
            {'type': 'Polygon', 'rounding_radius': 0.1,
             'vertices': [[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5]]}
        """
        return {'type': 'Polygon',
                'rounding_radius': self.rounding_radius,
                'vertices': self.vertices}


class ConvexPolyhedronShape(Shape):
    """Shape class for convex polyhedra.

    :param vertices:
        List of 3D vertex vectors.
    :type vertices:
        list
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, vertices, color=None):
        super(ConvexPolyhedronShape, self).__init__(
            shape_class='poly3d', color=color)
        self.vertices = vertices

    @property
    def pos_string(self):
        return "{} {} {} {}".format(
            self.shape_class,
            len(self.vertices),
            ' '.join((str(v) for xyz in self.vertices for v in xyz)),
            self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> ConvexPolyhedronShape([[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                                       [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]]).type_shape
            {'type': 'ConvexPolyhedron', 'rounding_radius': 0,
             'vertices': [[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                          [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]]}
        """
        return {'type': 'ConvexPolyhedron',
                'rounding_radius': 0,
                'vertices': self.vertices}


class ConvexPolyhedronUnionShape(Shape):
    """Shape class for unions of convex polyhedra.

    :param vertices:
        List of lists of 3D vertex vectors in particle coordinates (each
        polyhedron, each vertex).
    :type vertices:
        list
    :param centers:
        List of 3D polyhedra center vectors.
    :type centers:
        list
    :param orientations:
        Orientations of the polyhedra, as a list of quaternions.
    :type orientations:
        list
    :param colors:
        List of hexadecimal color strings in format :code:`RRGGBBAA` (default: :code:`None`).
    :type colors:
        list
    """

    def __init__(self, vertices, centers, orientations, colors=None):
        super(ConvexPolyhedronUnionShape, self).__init__(
            shape_class='poly3d_union', color='')
        self.vertices = vertices
        self.centers = centers
        self.orientations = orientations
        self.colors = colors

    @property
    def pos_string(self):
        shape_def = '{} {} '.format(self.shape_class, len(self.centers))
        for verts, p, q, c in zip(self.vertices, self.centers, self.orientations, self.colors):
            shape_def += '{0} '.format(len(verts))
            for v in verts:
                shape_def += '{0} {1} {2} '.format(*v)
            shape_def += '{0} {1} {2} '.format(*p)
            shape_def += '{0} {1} {2} {3} '.format(*q)
            shape_def += '{0} '.format(c)

        return shape_def


class ConvexSpheropolyhedronShape(Shape):
    """Shape class for a convex polyhedron extended by a rounding radius.

    :param vertices:
        List of 3D vertex vectors.
    :type vertices:
        list
    :param rounding_radius:
        Rounding radius applied to the spheropolyhedron (default: 0).
    :type rounding_radius:
        float
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, vertices, rounding_radius=0, color=None):
        super(ConvexSpheropolyhedronShape, self).__init__(
            shape_class='spoly3d', color=color)
        self.vertices = vertices
        self.rounding_radius = rounding_radius

    @property
    def pos_string(self):
        return "{} {} {} {} {}".format(
            self.shape_class,
            self.rounding_radius,
            len(self.vertices),
            ' '.join((str(v) for xyz in self.vertices for v in xyz)),
            self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> ConvexSpheropolyhedronShape([[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                                             [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]], 0.1).type_shape
            {'type': 'ConvexPolyhedron', 'rounding_radius': 0.1,
             'vertices': [[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                          [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]]}
        """
        return {'type': 'ConvexPolyhedron',
                'rounding_radius': self.rounding_radius,
                'vertices': self.vertices}


class GeneralPolyhedronShape(Shape):
    """Shape class for general polyhedra, such as arbitrary meshes.

    :param vertices:
        List of 3D vertex vectors.
    :type vertices:
        list
    :param faces:
        List of lists of integers representing vertex indices for each face.
    :type faces:
        list
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    :param facet_colors:
        List of hexadecimal color strings in format :code:`RRGGBBAA` for each facet (default: :code:`None`).
    :type facet_colors:
        list
    """

    def __init__(self, vertices, faces, color=None, facet_colors=None):
        super(GeneralPolyhedronShape, self).__init__(
            shape_class='polyV', color=color)
        self.vertices = vertices
        self.faces = faces
        self.facet_colors = facet_colors

    @property
    def pos_string(self):
        return "{} {} {} {} {} {}".format(
            self.shape_class,
            len(self.vertices),
            ' '.join((str(v) for xyz in self.vertices for v in xyz)),
            len(self.faces),
            ' '.join((str(fv) for f in self.faces for fv in [len(f)]+f)),
            self.color)

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> GeneralPolyhedronShape([[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                                        [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]]).type_shape
            {'type': 'Mesh',
             'vertices': [[0.5, 0.5, 0.5], [0.5, -0.5, -0.5],
                          [-0.5, 0.5, -0.5], [-0.5, -0.5, 0.5]],
             'indices': [[0, 1, 2], [0, 1, 3], [0, 2, 3], [1, 2, 3]]}
        """
        return {'type': 'Mesh',
                'vertices': self.vertices,
                'indices': self.faces}


class EllipsoidShape(Shape):
    """Shape class for ellipsoids of with principal axes a, b, and c.

    :param a:
        Principal axis a of the ellipsoid (radius in the x direction).
    :type a:
        float
    :param b:
        Principal axis b of the ellipsoid (radius in the y direction).
    :type b:
        float
    :param c:
        Principal axis c of the ellipsoid (radius in the z direction).
    :type c:
        float
    :param color:
        Hexadecimal color string in format :code:`RRGGBBAA` (default: :code:`None`).
    :type color:
        str
    """

    def __init__(self, a, b, c, color=None):
        super(EllipsoidShape, self).__init__(
            shape_class='ellipsoid', color=color)
        self.a = a
        self.b = b
        self.c = c

    @property
    def pos_string(self):
        return "{} {} {} {} {}".format(
            self.shape_class,
            self.a,
            self.b,
            self.c,
            self.color
        )

    @property
    @_json_sanitize
    def type_shape(self):
        """Shape as dictionary. Example:

            >>> EllipsoidShape(7.0, 5.0, 3.0).type_shape
            {'type': 'Ellipsoid',
             'a': 7.0,
             'b': 5.0,
             'c': 3.0}

        """
        return {'type': 'Ellipsoid',
                'a': self.a,
                'b': self.b,
                'c': self.c}


def _parse_type_shape(shape):
    """Parses a shape object from a dictionary.

    This method parses the `GSD Shape Visualization Specification
    <https://gsd.readthedocs.io/en/stable/shapes.html>`_, while including
    backwards compatibility with shape definitions that do not adhere to that
    specification but were previously supported by HOOMD's
    :code:`get_type_shapes()` methods.
    """

    if not shape:
        return FallbackShape('')

    type_name = shape['type'].lower()
    type_shape = None

    if type_name in ('sphere', 'disk'):
        # disk support is for backwards compatibility with get_type_shapes()
        # from HOOMD-blue < 2.7
        diameter = shape.get('diameter', 2*shape.get('rounding_radius', 0.5))
        orientable = shape.get('orientable', False)
        type_shape = SphereShape(diameter=diameter, orientable=orientable, color=None)
    elif type_name == 'ellipsoid':
        type_shape = EllipsoidShape(a=shape['a'], b=shape['b'], c=shape['c'], color=None)
    elif type_name == 'polygon':
        rounding_radius = shape.get('rounding_radius', 0)
        if rounding_radius == 0:
            type_shape = PolygonShape(vertices=shape['vertices'], color=None)
        else:
            type_shape = SpheropolygonShape(vertices=shape['vertices'],
                                            rounding_radius=rounding_radius,
                                            color=None)
    elif type_name == 'convexpolyhedron':
        rounding_radius = shape.get('rounding_radius', 0)
        if rounding_radius == 0:
            type_shape = ConvexPolyhedronShape(vertices=shape['vertices'], color=None)
        else:
            type_shape = ConvexSpheropolyhedronShape(vertices=shape['vertices'],
                                                     rounding_radius=rounding_radius,
                                                     color=None)
    elif type_name == 'mesh':
        type_shape = GeneralPolyhedronShape(vertices=shape['vertices'],
                                            faces=shape['indices'],
                                            facet_colors=shape['colors'],
                                            color=None)
    elif type_name == 'polyhedron':
        # polyhedron support is for backwards compatibility with
        # get_type_shapes() from HOOMD-blue < 2.7
        type_shape = GeneralPolyhedronShape(vertices=shape['vertices'],
                                            faces=shape['faces'],
                                            facet_colors=shape['colors'],
                                            color=None)
    elif type_name == 'sphereunion':
        type_shape = SphereUnionShape(diameters=shape['diameters'],
                                      centers=shape['centers'],
                                      color=None)

    if type_shape is None:
        logger.warning("Failed to parse shape definition: shape {} not supported. "
                       "Using fallback mode.".format(type_name))
        type_shape = FallbackShape(json.dumps(shape))

    return type_shape
